from __future__ import annotations

from ..._utils import is_dict, is_list


def accumulate_delta(acc: dict[object, object], delta: dict[object, object]) -> dict[object, object]:
    for key, delta_value in delta.items():
        if key not in acc:
            acc[key] = delta_value
            continue

        acc_value = acc[key]
        if acc_value is None:
            acc[key] = delta_value
            continue

        # the `index` property is used in arrays of objects so it should
        # not be accumulated like other values e.g.
        # [{'foo': 'bar', 'index': 0}]
        #
        # the same applies to `type` properties as they're used for
        # discriminated unions
        if key == "index" or key == "type" or key == "role":
            acc[key] = delta_value
            continue

        if isinstance(acc_value, str) and isinstance(delta_value, str):
            acc_value += delta_value
        elif isinstance(acc_value, (int, float)) and isinstance(delta_value, (int, float)):
            acc_value += delta_value
        elif is_dict(acc_value) and is_dict(delta_value):
            acc_value = accumulate_delta(acc_value, delta_value)
        elif is_list(acc_value) and is_list(delta_value):
            # for lists of non-dictionary items we'll only ever get new entries
            # in the array, existing entries will never be changed
            if all(isinstance(x, (str, int, float)) for x in acc_value):
                acc_value.extend(delta_value)
                continue

            for delta_entry in delta_value:
                if not is_dict(delta_entry):
                    raise TypeError(f"Unexpected list delta entry is not a dictionary: {delta_entry}")

                try:
                    index = delta_entry["index"]
                except KeyError as exc:
                    raise RuntimeError(f"Expected list delta entry to have an `index` key; {delta_entry}") from exc

                if not isinstance(index, int):
                    raise TypeError(f"Unexpected, list delta entry `index` value is not an integer; {index}")

                try:
                    acc_entry = acc_value[index]
                except IndexError:
                    acc_value.insert(index, delta_entry)
                else:
                    if not is_dict(acc_entry):
                        raise TypeError("not handled yet")

                    acc_value[index] = accumulate_delta(acc_entry, delta_entry)

        acc[key] = acc_value

    return acc



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .tools import (
    ToolsResource,
    AsyncToolsResource,
    ToolsResourceWithRawResponse,
    AsyncToolsResourceWithRawResponse,
    ToolsResourceWithStreamingResponse,
    AsyncToolsResourceWithStreamingResponse,
)
from .comprehend import (
    ComprehendResource,
    AsyncComprehendResource,
    ComprehendResourceWithRawResponse,
    AsyncComprehendResourceWithRawResponse,
    ComprehendResourceWithStreamingResponse,
    AsyncComprehendResourceWithStreamingResponse,
)

__all__ = [
    "ComprehendResource",
    "AsyncComprehendResource",
    "ComprehendResourceWithRawResponse",
    "AsyncComprehendResourceWithRawResponse",
    "ComprehendResourceWithStreamingResponse",
    "AsyncComprehendResourceWithStreamingResponse",
    "ToolsResource",
    "AsyncToolsResource",
    "ToolsResourceWithRawResponse",
    "AsyncToolsResourceWithRawResponse",
    "ToolsResourceWithStreamingResponse",
    "AsyncToolsResourceWithStreamingResponse",
]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .._models import BaseModel

__all__ = ["TranslationResponse"]


class TranslationResponse(BaseModel):
    data: str
    """The result of the translation."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional
from datetime import datetime
from typing_extensions import Literal

from .._models import BaseModel

__all__ = ["Graph", "FileStatus"]


class FileStatus(BaseModel):
    completed: int
    """The number of files that have been successfully processed."""

    failed: int
    """The number of files that failed to process."""

    in_progress: int
    """The number of files currently being processed."""

    total: int
    """The total number of files associated with the Knowledge Graph."""


class Graph(BaseModel):
    id: str
    """The unique identifier of the Knowledge Graph."""

    created_at: datetime
    """The timestamp when the Knowledge Graph was created."""

    file_status: FileStatus

    name: str
    """The name of the Knowledge Graph."""

    type: Literal["manual", "connector"]
    """
    The type of Knowledge Graph, either `manual` (files are uploaded via UI or API)
    or `connector` (files are uploaded via a connector).
    """

    description: Optional[str] = None
    """A description of the Knowledge Graph."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import List
from typing_extensions import Required, TypedDict

__all__ = ["GraphUpdateParams"]


class GraphUpdateParams(TypedDict, total=False):
    graph_ids: Required[List[str]]
    """A list of Knowledge Graph IDs to associate with the application.

    Note that this will replace the existing list of Knowledge Graphs associated
    with the application, not add to it.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import List, Union, Iterable, Optional
from typing_extensions import Literal, Required, TypeAlias, TypedDict

from .shared_params.tool_call import ToolCall
from .shared_params.graph_data import GraphData
from .shared_params.tool_param import ToolParam
from .shared_params.tool_choice_string import ToolChoiceString
from .shared_params.tool_choice_json_object import ToolChoiceJsonObject

__all__ = [
    "ChatChatParamsBase",
    "Message",
    "ResponseFormat",
    "StreamOptions",
    "ToolChoice",
    "ChatChatParamsNonStreaming",
    "ChatChatParamsStreaming",
]


class ChatChatParamsBase(TypedDict, total=False):
    messages: Required[Iterable[Message]]
    """
    An array of message objects that form the conversation history or context for
    the model to respond to. The array must contain at least one message.
    """

    model: Required[str]
    """
    The [ID of the model](https://dev.writer.com/home/models) to use for creating
    the chat completion. Supports `palmyra-x5`, `palmyra-x4`, `palmyra-fin`,
    `palmyra-med`, `palmyra-creative`, and `palmyra-x-003-instruct`.
    """

    logprobs: bool
    """Specifies whether to return log probabilities of the output tokens."""

    max_tokens: int
    """
    Defines the maximum number of tokens (words and characters) that the model can
    generate in the response. The default value is set to 16, but it can be adjusted
    to allow for longer or shorter responses as needed.
    """

    n: int
    """
    Specifies the number of completions (responses) to generate from the model in a
    single request. This parameter allows for generating multiple responses,
    offering a variety of potential replies from which to choose.
    """

    response_format: ResponseFormat
    """
    The response format to use for the chat completion, available with `palmyra-x4`
    and `palmyra-x5`.

    `text` is the default response format. [JSON Schema](https://json-schema.org/)
    is supported for structured responses. If you specify `json_schema`, you must
    also provide a `json_schema` object.
    """

    stop: Union[List[str], str]
    """
    A token or sequence of tokens that, when generated, will cause the model to stop
    producing further content. This can be a single token or an array of tokens,
    acting as a signal to end the output.
    """

    stream_options: StreamOptions
    """Additional options for streaming."""

    temperature: float
    """Controls the randomness or creativity of the model's responses.

    A higher temperature results in more varied and less predictable text, while a
    lower temperature produces more deterministic and conservative outputs.
    """

    tool_choice: ToolChoice
    """
    Configure how the model will call functions: `auto` will allow the model to
    automatically choose the best tool, `none` disables tool calling. You can also
    pass a specific previously defined function.
    """

    tools: Iterable[ToolParam]
    """
    An array containing tool definitions for tools that the model can use to
    generate responses. The tool definitions use JSON schema. You can define your
    own functions or use one of the built-in `graph`, `llm`, `translation`, or
    `vision` tools. Note that you can only use one built-in tool type in the array
    (only one of `graph`, `llm`, `translation`, or `vision`). You can pass multiple
    [custom tools](https://dev.writer.com/api-guides/tool-calling) of type
    `function` in the same request.
    """

    top_p: float
    """
    Sets the threshold for "nucleus sampling," a technique to focus the model's
    token generation on the most likely subset of tokens. Only tokens with
    cumulative probability above this threshold are considered, controlling the
    trade-off between creativity and coherence.
    """


class Message(TypedDict, total=False):
    role: Required[Literal["user", "assistant", "system", "tool"]]
    """The role of the chat message.

    You can provide a system prompt by setting the role to `system`, or specify that
    a message is the result of a
    [tool call](https://dev.writer.com/api-guides/tool-calling) by setting the role
    to `tool`.
    """

    content: Optional[str]

    graph_data: Optional[GraphData]

    name: Optional[str]

    refusal: Optional[str]

    tool_call_id: Optional[str]

    tool_calls: Optional[Iterable[ToolCall]]


class ResponseFormat(TypedDict, total=False):
    type: Required[Literal["text", "json_schema"]]
    """The type of response format to use."""

    json_schema: object
    """The JSON schema to use for the response format."""


class StreamOptions(TypedDict, total=False):
    include_usage: Required[bool]
    """Indicate whether to include usage information."""


ToolChoice: TypeAlias = Union[ToolChoiceString, ToolChoiceJsonObject]


class ChatChatParamsNonStreaming(ChatChatParamsBase, total=False):
    stream: Literal[False]
    """
    Indicates whether the response should be streamed incrementally as it is
    generated or only returned once fully complete. Streaming can be useful for
    providing real-time feedback in interactive applications.
    """


class ChatChatParamsStreaming(ChatChatParamsBase):
    stream: Required[Literal[True]]
    """
    Indicates whether the response should be streamed incrementally as it is
    generated or only returned once fully complete. Streaming can be useful for
    providing real-time feedback in interactive applications.
    """


ChatChatParams = Union[ChatChatParamsNonStreaming, ChatChatParamsStreaming]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Optional
from typing_extensions import Literal

from .source import Source
from ..._models import BaseModel

__all__ = ["GraphData", "Subquery"]


class Subquery(BaseModel):
    answer: str
    """The answer to the subquery."""

    query: str
    """The subquery that was asked."""

    sources: List[Optional[Source]]


class GraphData(BaseModel):
    sources: Optional[List[Optional[Source]]] = None

    status: Optional[Literal["processing", "finished"]] = None

    subqueries: Optional[List[Optional[Subquery]]] = None



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from datetime import datetime
from typing_extensions import Literal

from ..._models import BaseModel

__all__ = ["JobRetryResponse"]


class JobRetryResponse(BaseModel):
    id: str
    """The unique identifier for the async job created."""

    created_at: datetime
    """The timestamp when the job was created."""

    status: Literal["in_progress", "failed", "completed"]
    """The status of the job."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Iterable, Optional
from typing_extensions import Literal, Required, TypedDict

from .source import Source

__all__ = ["GraphData", "Subquery"]


class Subquery(TypedDict, total=False):
    answer: Required[str]
    """The answer to the subquery."""

    query: Required[str]
    """The subquery that was asked."""

    sources: Required[Iterable[Optional[Source]]]


class GraphData(TypedDict, total=False):
    sources: Iterable[Optional[Source]]

    status: Optional[Literal["processing", "finished"]]

    subqueries: Iterable[Optional[Subquery]]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Union, Optional
from typing_extensions import Literal, Annotated, TypeAlias

from ..._utils import PropertyInfo
from ..._models import BaseModel
from .function_definition import FunctionDefinition

__all__ = [
    "ToolParam",
    "FunctionTool",
    "GraphTool",
    "GraphToolFunction",
    "LlmTool",
    "LlmToolFunction",
    "TranslationTool",
    "TranslationToolFunction",
    "VisionTool",
    "VisionToolFunction",
    "VisionToolFunctionVariable",
]


class FunctionTool(BaseModel):
    function: FunctionDefinition
    """A tool that uses a custom function."""

    type: Literal["function"]
    """The type of tool."""


class GraphToolFunction(BaseModel):
    graph_ids: List[str]
    """An array of graph IDs to use in the tool."""

    subqueries: bool
    """Boolean to indicate whether to include subqueries in the response."""

    description: Optional[str] = None
    """A description of the graph content."""


class GraphTool(BaseModel):
    function: GraphToolFunction
    """A tool that uses Knowledge Graphs as context for responses."""

    type: Literal["graph"]
    """The type of tool."""


class LlmToolFunction(BaseModel):
    description: str
    """A description of the model to use."""

    model: str
    """The model to use."""


class LlmTool(BaseModel):
    function: LlmToolFunction
    """A tool that uses another Writer model to generate a response."""

    type: Literal["llm"]
    """The type of tool."""


class TranslationToolFunction(BaseModel):
    formality: bool
    """Whether to use formal or informal language in the translation.

    See the
    [list of languages that support formality](https://dev.writer.com/api-guides/api-reference/translation-api/language-support#formality).
    If the language does not support formality, this parameter is ignored.
    """

    length_control: bool
    """Whether to control the length of the translated text.

    See the
    [list of languages that support length control](https://dev.writer.com/api-guides/api-reference/translation-api/language-support#length-control).
    If the language does not support length control, this parameter is ignored.
    """

    mask_profanity: bool
    """Whether to mask profane words in the translated text.

    See the
    [list of languages that do not support profanity masking](https://dev.writer.com/api-guides/api-reference/translation-api/language-support#profanity-masking).
    If the language does not support profanity masking, this parameter is ignored.
    """

    model: Literal["palmyra-translate"]
    """The model to use for translation."""

    source_language_code: Optional[str] = None
    """Optional.

    The [ISO-639-1](https://en.wikipedia.org/wiki/List_of_ISO_639_language_codes)
    language code of the original text to translate. For example, `en` for English,
    `zh` for Chinese, `fr` for French, `es` for Spanish. If the language has a
    variant, the code appends the two-digit
    [ISO-3166 country code](https://en.wikipedia.org/wiki/List_of_ISO_3166_country_codes).
    If you do not provide a language code, the LLM detects the language of the text.
    """

    target_language_code: Optional[str] = None
    """Optional.

    The [ISO-639-1](https://en.wikipedia.org/wiki/List_of_ISO_639_language_codes)
    language code of the target language for the translation. For example, `en` for
    English, `zh` for Chinese, `fr` for French, `es` for Spanish. If the language
    has a variant, the code appends the two-digit
    [ISO-3166 country code](https://en.wikipedia.org/wiki/List_of_ISO_3166_country_codes).
    If you do not provide a language code, the LLM uses the content of the chat
    message to determine the target language.
    """


class TranslationTool(BaseModel):
    function: TranslationToolFunction
    """A tool that uses Palmyra Translate to translate text."""

    type: Literal["translation"]
    """The type of tool."""


class VisionToolFunctionVariable(BaseModel):
    file_id: str
    """The File ID of the image to analyze.

    The file must be uploaded to the Writer platform before you use it with the
    Vision tool.
    """

    name: str
    """The name of the file variable.

    You must reference this name in the `message.content` field of the request to
    the chat completions endpoint. Use double curly braces (`{{}}`) to reference the
    file. For example,
    `Describe the difference between the image {{image_1}} and the image {{image_2}}`.
    """


class VisionToolFunction(BaseModel):
    model: Literal["palmyra-vision"]
    """The model to use for image analysis."""

    variables: List[VisionToolFunctionVariable]


class VisionTool(BaseModel):
    function: VisionToolFunction
    """A tool that uses Palmyra Vision to analyze images."""

    type: Literal["vision"]
    """The type of tool."""


ToolParam: TypeAlias = Annotated[
    Union[FunctionTool, GraphTool, LlmTool, TranslationTool, VisionTool], PropertyInfo(discriminator="type")
]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional
from typing_extensions import Literal

from ..._models import BaseModel

__all__ = ["ToolCall", "Function"]


class Function(BaseModel):
    arguments: str

    name: Optional[str] = None


class ToolCall(BaseModel):
    id: str

    function: Function

    type: Literal["function"]

    index: Optional[int] = None



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import List, Iterable
from typing_extensions import Required, TypedDict

__all__ = ["JobCreateParams", "Input"]


class JobCreateParams(TypedDict, total=False):
    inputs: Required[Iterable[Input]]
    """A list of input objects to generate content for."""


class Input(TypedDict, total=False):
    id: Required[str]
    """The unique identifier for the input field from the application.

    All input types from the No-code application are supported (i.e. Text input,
    Dropdown, File upload, Image input). The identifier should be the name of the
    input type.
    """

    value: Required[List[str]]
    """The value for the input field.

    If the input type is "File upload", you must pass the `file_id` of an uploaded
    file. You cannot pass a file object directly. See the
    [file upload endpoint](https://dev.writer.com/api-guides/api-reference/file-api/upload-files)
    for instructions on uploading files or the
    [list files endpoint](https://dev.writer.com/api-guides/api-reference/file-api/get-all-files)
    for how to see a list of uploaded files and their IDs.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Union, Optional
from datetime import datetime
from typing_extensions import Literal, TypeAlias

from .._models import BaseModel

__all__ = [
    "ApplicationRetrieveResponse",
    "Input",
    "InputOptions",
    "InputOptionsApplicationInputDropdownOptions",
    "InputOptionsApplicationInputFileOptions",
    "InputOptionsApplicationInputMediaOptions",
    "InputOptionsApplicationInputTextOptions",
]


class InputOptionsApplicationInputDropdownOptions(BaseModel):
    list: List[str]
    """List of available options in the dropdown menu."""


class InputOptionsApplicationInputFileOptions(BaseModel):
    file_types: List[str]
    """List of allowed file extensions."""

    max_file_size_mb: int
    """Maximum file size allowed in megabytes."""

    max_files: int
    """Maximum number of files that can be uploaded."""

    max_word_count: int
    """Maximum number of words allowed in text files."""


class InputOptionsApplicationInputMediaOptions(BaseModel):
    file_types: List[str]
    """List of allowed media file types."""

    max_image_size_mb: int
    """Maximum media file size allowed in megabytes."""


class InputOptionsApplicationInputTextOptions(BaseModel):
    max_fields: int
    """Maximum number of text fields allowed."""

    min_fields: int
    """Minimum number of text fields required."""


InputOptions: TypeAlias = Union[
    InputOptionsApplicationInputDropdownOptions,
    InputOptionsApplicationInputFileOptions,
    InputOptionsApplicationInputMediaOptions,
    InputOptionsApplicationInputTextOptions,
]


class Input(BaseModel):
    input_type: Literal["text", "dropdown", "file", "media"]
    """Type of input field determining its behavior and validation rules."""

    name: str
    """Identifier for the input field."""

    required: bool
    """Indicates if this input field is mandatory."""

    description: Optional[str] = None
    """Human-readable description of the input field's purpose."""

    options: Optional[InputOptions] = None
    """Type-specific configuration options for input fields."""


class ApplicationRetrieveResponse(BaseModel):
    id: str
    """Unique identifier for the application."""

    created_at: datetime
    """Timestamp when the application was created."""

    inputs: List[Input]
    """List of input configurations for the application."""

    name: str
    """Display name of the application."""

    status: Literal["deployed", "draft"]
    """Current deployment status of the application.

    Note: currently only `deployed` applications are returned.
    """

    type: Literal["generation"]
    """The type of no-code application."""

    updated_at: datetime
    """Timestamp when the application was last updated."""

    last_deployed_at: Optional[datetime] = None
    """Timestamp when the application was last deployed."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Iterable
from typing_extensions import Literal

import httpx

from ..._types import NOT_GIVEN, Body, Query, Headers, NotGiven
from ..._utils import maybe_transform, async_maybe_transform
from ..._compat import cached_property
from ..._resource import SyncAPIResource, AsyncAPIResource
from ..._response import (
    to_raw_response_wrapper,
    to_streamed_response_wrapper,
    async_to_raw_response_wrapper,
    async_to_streamed_response_wrapper,
)
from ...pagination import SyncApplicationJobsOffset, AsyncApplicationJobsOffset
from ..._base_client import AsyncPaginator, make_request_options
from ...types.applications import job_list_params, job_create_params
from ...types.applications.job_retry_response import JobRetryResponse
from ...types.applications.job_create_response import JobCreateResponse
from ...types.applications.application_generate_async_response import ApplicationGenerateAsyncResponse

__all__ = ["JobsResource", "AsyncJobsResource"]


class JobsResource(SyncAPIResource):
    @cached_property
    def with_raw_response(self) -> JobsResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/writer/writer-python#accessing-raw-response-data-eg-headers
        """
        return JobsResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> JobsResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/writer/writer-python#with_streaming_response
        """
        return JobsResourceWithStreamingResponse(self)

    def create(
        self,
        application_id: str,
        *,
        inputs: Iterable[job_create_params.Input],
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> JobCreateResponse:
        """
        Generate content asynchronously from an existing no-code agent (formerly called
        no-code applications) with inputs.

        Args:
          inputs: A list of input objects to generate content for.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not application_id:
            raise ValueError(f"Expected a non-empty value for `application_id` but received {application_id!r}")
        return self._post(
            f"/v1/applications/{application_id}/jobs",
            body=maybe_transform({"inputs": inputs}, job_create_params.JobCreateParams),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=JobCreateResponse,
        )

    def retrieve(
        self,
        job_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> ApplicationGenerateAsyncResponse:
        """
        Retrieves a single job created via the Async API.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not job_id:
            raise ValueError(f"Expected a non-empty value for `job_id` but received {job_id!r}")
        return self._get(
            f"/v1/applications/jobs/{job_id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=ApplicationGenerateAsyncResponse,
        )

    def list(
        self,
        application_id: str,
        *,
        limit: int | NotGiven = NOT_GIVEN,
        offset: int | NotGiven = NOT_GIVEN,
        status: Literal["in_progress", "failed", "completed"] | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> SyncApplicationJobsOffset[ApplicationGenerateAsyncResponse]:
        """
        Retrieve all jobs created via the async API, linked to the provided application
        ID (or alias).

        Args:
          limit: The pagination limit for retrieving the jobs.

          offset: The pagination offset for retrieving the jobs.

          status: The status of the job.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not application_id:
            raise ValueError(f"Expected a non-empty value for `application_id` but received {application_id!r}")
        return self._get_api_list(
            f"/v1/applications/{application_id}/jobs",
            page=SyncApplicationJobsOffset[ApplicationGenerateAsyncResponse],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "limit": limit,
                        "offset": offset,
                        "status": status,
                    },
                    job_list_params.JobListParams,
                ),
            ),
            model=ApplicationGenerateAsyncResponse,
        )

    def retry(
        self,
        job_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> JobRetryResponse:
        """
        Re-triggers the async execution of a single job previously created via the Async
        api and terminated in error.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not job_id:
            raise ValueError(f"Expected a non-empty value for `job_id` but received {job_id!r}")
        return self._post(
            f"/v1/applications/jobs/{job_id}/retry",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=JobRetryResponse,
        )


class AsyncJobsResource(AsyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AsyncJobsResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/writer/writer-python#accessing-raw-response-data-eg-headers
        """
        return AsyncJobsResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncJobsResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/writer/writer-python#with_streaming_response
        """
        return AsyncJobsResourceWithStreamingResponse(self)

    async def create(
        self,
        application_id: str,
        *,
        inputs: Iterable[job_create_params.Input],
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> JobCreateResponse:
        """
        Generate content asynchronously from an existing no-code agent (formerly called
        no-code applications) with inputs.

        Args:
          inputs: A list of input objects to generate content for.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not application_id:
            raise ValueError(f"Expected a non-empty value for `application_id` but received {application_id!r}")
        return await self._post(
            f"/v1/applications/{application_id}/jobs",
            body=await async_maybe_transform({"inputs": inputs}, job_create_params.JobCreateParams),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=JobCreateResponse,
        )

    async def retrieve(
        self,
        job_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> ApplicationGenerateAsyncResponse:
        """
        Retrieves a single job created via the Async API.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not job_id:
            raise ValueError(f"Expected a non-empty value for `job_id` but received {job_id!r}")
        return await self._get(
            f"/v1/applications/jobs/{job_id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=ApplicationGenerateAsyncResponse,
        )

    def list(
        self,
        application_id: str,
        *,
        limit: int | NotGiven = NOT_GIVEN,
        offset: int | NotGiven = NOT_GIVEN,
        status: Literal["in_progress", "failed", "completed"] | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> AsyncPaginator[ApplicationGenerateAsyncResponse, AsyncApplicationJobsOffset[ApplicationGenerateAsyncResponse]]:
        """
        Retrieve all jobs created via the async API, linked to the provided application
        ID (or alias).

        Args:
          limit: The pagination limit for retrieving the jobs.

          offset: The pagination offset for retrieving the jobs.

          status: The status of the job.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not application_id:
            raise ValueError(f"Expected a non-empty value for `application_id` but received {application_id!r}")
        return self._get_api_list(
            f"/v1/applications/{application_id}/jobs",
            page=AsyncApplicationJobsOffset[ApplicationGenerateAsyncResponse],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "limit": limit,
                        "offset": offset,
                        "status": status,
                    },
                    job_list_params.JobListParams,
                ),
            ),
            model=ApplicationGenerateAsyncResponse,
        )

    async def retry(
        self,
        job_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> JobRetryResponse:
        """
        Re-triggers the async execution of a single job previously created via the Async
        api and terminated in error.

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not job_id:
            raise ValueError(f"Expected a non-empty value for `job_id` but received {job_id!r}")
        return await self._post(
            f"/v1/applications/jobs/{job_id}/retry",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=JobRetryResponse,
        )


class JobsResourceWithRawResponse:
    def __init__(self, jobs: JobsResource) -> None:
        self._jobs = jobs

        self.create = to_raw_response_wrapper(
            jobs.create,
        )
        self.retrieve = to_raw_response_wrapper(
            jobs.retrieve,
        )
        self.list = to_raw_response_wrapper(
            jobs.list,
        )
        self.retry = to_raw_response_wrapper(
            jobs.retry,
        )


class AsyncJobsResourceWithRawResponse:
    def __init__(self, jobs: AsyncJobsResource) -> None:
        self._jobs = jobs

        self.create = async_to_raw_response_wrapper(
            jobs.create,
        )
        self.retrieve = async_to_raw_response_wrapper(
            jobs.retrieve,
        )
        self.list = async_to_raw_response_wrapper(
            jobs.list,
        )
        self.retry = async_to_raw_response_wrapper(
            jobs.retry,
        )


class JobsResourceWithStreamingResponse:
    def __init__(self, jobs: JobsResource) -> None:
        self._jobs = jobs

        self.create = to_streamed_response_wrapper(
            jobs.create,
        )
        self.retrieve = to_streamed_response_wrapper(
            jobs.retrieve,
        )
        self.list = to_streamed_response_wrapper(
            jobs.list,
        )
        self.retry = to_streamed_response_wrapper(
            jobs.retry,
        )


class AsyncJobsResourceWithStreamingResponse:
    def __init__(self, jobs: AsyncJobsResource) -> None:
        self._jobs = jobs

        self.create = async_to_streamed_response_wrapper(
            jobs.create,
        )
        self.retrieve = async_to_streamed_response_wrapper(
            jobs.retrieve,
        )
        self.list = async_to_streamed_response_wrapper(
            jobs.list,
        )
        self.retry = async_to_streamed_response_wrapper(
            jobs.retry,
        )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Any, List, Generic, TypeVar, Optional, cast
from typing_extensions import Protocol, override, runtime_checkable

from pydantic import Field as FieldInfo

from ._models import BaseModel
from ._base_client import BasePage, PageInfo, BaseSyncPage, BaseAsyncPage

__all__ = [
    "SyncCursorPage",
    "AsyncCursorPage",
    "ApplicationJobsOffsetPagination",
    "SyncApplicationJobsOffset",
    "AsyncApplicationJobsOffset",
]

_T = TypeVar("_T")


@runtime_checkable
class CursorPageItem(Protocol):
    id: Optional[str]


class SyncCursorPage(BaseSyncPage[_T], BasePage[_T], Generic[_T]):
    data: List[_T]
    has_more: bool

    @override
    def _get_page_items(self) -> List[_T]:
        data = self.data
        if not data:
            return []
        return data

    @override
    def has_next_page(self) -> bool:
        has_more = self.has_more
        return has_more and super().has_next_page()

    @override
    def next_page_info(self) -> Optional[PageInfo]:
        is_forwards = not self._options.params.get("before", False)

        data = self.data
        if not data:
            return None

        if is_forwards:
            item = cast(Any, data[-1])
            if not isinstance(item, CursorPageItem) or item.id is None:
                # TODO emit warning log
                return None

            return PageInfo(params={"after": item.id})
        else:
            item = cast(Any, self.data[0])
            if not isinstance(item, CursorPageItem) or item.id is None:
                # TODO emit warning log
                return None

            return PageInfo(params={"before": item.id})


class AsyncCursorPage(BaseAsyncPage[_T], BasePage[_T], Generic[_T]):
    data: List[_T]
    has_more: bool

    @override
    def _get_page_items(self) -> List[_T]:
        data = self.data
        if not data:
            return []
        return data

    @override
    def has_next_page(self) -> bool:
        has_more = self.has_more
        return has_more and super().has_next_page()

    @override
    def next_page_info(self) -> Optional[PageInfo]:
        is_forwards = not self._options.params.get("before", False)

        data = self.data
        if not data:
            return None

        if is_forwards:
            item = cast(Any, data[-1])
            if not isinstance(item, CursorPageItem) or item.id is None:
                # TODO emit warning log
                return None

            return PageInfo(params={"after": item.id})
        else:
            item = cast(Any, self.data[0])
            if not isinstance(item, CursorPageItem) or item.id is None:
                # TODO emit warning log
                return None

            return PageInfo(params={"before": item.id})


class ApplicationJobsOffsetPagination(BaseModel):
    limit: Optional[int] = None

    offset: Optional[int] = None


class SyncApplicationJobsOffset(BaseSyncPage[_T], BasePage[_T], Generic[_T]):
    result: List[_T]
    total_count: Optional[int] = FieldInfo(alias="totalCount", default=None)
    pagination: Optional[ApplicationJobsOffsetPagination] = None

    @override
    def _get_page_items(self) -> List[_T]:
        result = self.result
        if not result:
            return []
        return result

    @override
    def next_page_info(self) -> Optional[PageInfo]:
        offset = None
        if self.pagination is not None:
            if self.pagination.offset is not None:
                offset = self.pagination.offset
        if offset is None:
            return None  # type: ignore[unreachable]

        length = len(self._get_page_items())
        current_count = offset + length

        total_count = self.total_count
        if total_count is None:
            return None

        if current_count < total_count:
            return PageInfo(params={"offset": current_count})

        return None


class AsyncApplicationJobsOffset(BaseAsyncPage[_T], BasePage[_T], Generic[_T]):
    result: List[_T]
    total_count: Optional[int] = FieldInfo(alias="totalCount", default=None)
    pagination: Optional[ApplicationJobsOffsetPagination] = None

    @override
    def _get_page_items(self) -> List[_T]:
        result = self.result
        if not result:
            return []
        return result

    @override
    def next_page_info(self) -> Optional[PageInfo]:
        offset = None
        if self.pagination is not None:
            if self.pagination.offset is not None:
                offset = self.pagination.offset
        if offset is None:
            return None  # type: ignore[unreachable]

        length = len(self._get_page_items())
        current_count = offset + length

        total_count = self.total_count
        if total_count is None:
            return None

        if current_count < total_count:
            return PageInfo(params={"offset": current_count})

        return None



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .._models import BaseModel

__all__ = ["FileDeleteResponse"]


class FileDeleteResponse(BaseModel):
    id: str
    """A unique identifier of the deleted file."""

    deleted: bool
    """Indicates whether the file was successfully deleted."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import Required, TypedDict

__all__ = ["ToolAIDetectParams"]


class ToolAIDetectParams(TypedDict, total=False):
    input: Required[str]
    """The content to determine if it is AI- or human-generated.

    Content must have at least 350 characters.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Generic, TypeVar, Optional

from .._models import GenericModel
from .chat_completion import ChatCompletion
from .chat_completion_choice import ChatCompletionChoice
from .chat_completion_message import ChatCompletionMessage
from .parsed_function_tool_call import ParsedFunctionToolCall

__all__ = ["ParsedChatCompletion", "ParsedChatCompletionChoice", "ParsedChatCompletionMessage"]


ContentType = TypeVar("ContentType")


# we need to disable this check because we're overriding properties
# with subclasses of their types which is technically unsound as
# properties can be mutated.
# pyright: reportIncompatibleVariableOverride=false


class ParsedChatCompletionMessage(ChatCompletionMessage, GenericModel, Generic[ContentType]):
    parsed: Optional[ContentType] = None
    """The auto-parsed message contents"""

    tool_calls: Optional[List[ParsedFunctionToolCall]] = None  # type: ignore[assignment]
    """The tool calls generated by the model, such as function calls."""


class ParsedChatCompletionChoice(ChatCompletionChoice, GenericModel, Generic[ContentType]):
    message: ParsedChatCompletionMessage[ContentType]
    """A chat completion message generated by the model."""


class ParsedChatCompletion(ChatCompletion, GenericModel, Generic[ContentType]):
    choices: List[ParsedChatCompletionChoice[ContentType]]  # type: ignore[assignment]
    """A list of chat completion choices.

    Can be more than one if `n` is greater than 1.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import Literal, Required, TypedDict

__all__ = ["ToolParsePdfParams"]


class ToolParsePdfParams(TypedDict, total=False):
    format: Required[Literal["text", "markdown"]]
    """The format into which the PDF content should be converted."""



from ._types import (
    ParsedChatCompletionSnapshot as ParsedChatCompletionSnapshot,
    ParsedChatCompletionChoiceSnapshot as ParsedChatCompletionChoiceSnapshot,
    ParsedChatCompletionMessageSnapshot as ParsedChatCompletionMessageSnapshot,
)
from ._events import (
    ChunkEvent as ChunkEvent,
    ContentDoneEvent as ContentDoneEvent,
    RefusalDoneEvent as RefusalDoneEvent,
    ContentDeltaEvent as ContentDeltaEvent,
    RefusalDeltaEvent as RefusalDeltaEvent,
    LogprobsContentDoneEvent as LogprobsContentDoneEvent,
    LogprobsRefusalDoneEvent as LogprobsRefusalDoneEvent,
    ChatCompletionStreamEvent as ChatCompletionStreamEvent,
    LogprobsContentDeltaEvent as LogprobsContentDeltaEvent,
    LogprobsRefusalDeltaEvent as LogprobsRefusalDeltaEvent,
    ParsedChatCompletionSnapshot as ParsedChatCompletionSnapshot,
    FunctionToolCallArgumentsDoneEvent as FunctionToolCallArgumentsDoneEvent,
    FunctionToolCallArgumentsDeltaEvent as FunctionToolCallArgumentsDeltaEvent,
)
from ._completions import (
    ChatCompletionStream as ChatCompletionStream,
    AsyncChatCompletionStream as AsyncChatCompletionStream,
    ChatCompletionStreamManager as ChatCompletionStreamManager,
    AsyncChatCompletionStreamManager as AsyncChatCompletionStreamManager,
)



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional

from .shared.tool_call import Function, ToolCall

__all__ = ["ParsedFunctionToolCall", "ParsedFunction"]

# we need to disable this check because we're overriding properties
# with subclasses of their types which is technically unsound as
# properties can be mutated.
# pyright: reportIncompatibleVariableOverride=false


class ParsedFunction(Function):
    parsed_arguments: Optional[object] = None
    """
    The arguments to call the function with.

    If you used `writerai.pydantic_function_tool()` then this will be an
    instance of the given `BaseModel`.

    Otherwise, this will be the parsed JSON arguments.
    """


class ParsedFunctionToolCall(ToolCall):
    function: ParsedFunction
    """The function that the model called."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .source import Source as Source
from .tool_call import ToolCall as ToolCall
from .graph_data import GraphData as GraphData
from .tool_param import ToolParam as ToolParam
from .function_params import FunctionParams as FunctionParams
from .response_format import ResponseFormat as ResponseFormat
from .tool_choice_string import ToolChoiceString as ToolChoiceString
from .function_definition import FunctionDefinition as FunctionDefinition
from .response_format_text import ResponseFormatText as ResponseFormatText
from .tool_choice_json_object import ToolChoiceJsonObject as ToolChoiceJsonObject
from .response_format_json_object import ResponseFormatJSONObject as ResponseFormatJSONObject
from .response_format_json_schema import ResponseFormatJSONSchema as ResponseFormatJSONSchema



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional

from .._models import BaseModel

__all__ = ["ApplicationGenerateContentResponse"]


class ApplicationGenerateContentResponse(BaseModel):
    suggestion: str
    """The response from the model specified in the application."""

    title: Optional[str] = None
    """The name of the output field."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional
from datetime import datetime

from .._models import BaseModel

__all__ = ["GraphUpdateResponse"]


class GraphUpdateResponse(BaseModel):
    id: str
    """A unique identifier of the Knowledge Graph."""

    created_at: datetime
    """The timestamp when the Knowledge Graph was created."""

    name: str
    """The name of the Knowledge Graph (max 255 characters)."""

    description: Optional[str] = None
    """A description of the Knowledge Graph (max 255 characters)."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Dict, List

from ..._models import BaseModel
from .error_message import ErrorMessage

__all__ = ["ErrorObject"]


class ErrorObject(BaseModel):
    errors: List[ErrorMessage]

    extras: Dict[str, object]

    tpe: str



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .source import Source as Source
from .logprobs import Logprobs as Logprobs
from .tool_call import ToolCall as ToolCall
from .graph_data import GraphData as GraphData
from .tool_param import ToolParam as ToolParam
from .error_object import ErrorObject as ErrorObject
from .error_message import ErrorMessage as ErrorMessage
from .logprobs_token import LogprobsToken as LogprobsToken
from .function_params import FunctionParams as FunctionParams
from .tool_choice_string import ToolChoiceString as ToolChoiceString
from .function_definition import FunctionDefinition as FunctionDefinition
from .tool_call_streaming import ToolCallStreaming as ToolCallStreaming
from .tool_choice_json_object import ToolChoiceJsonObject as ToolChoiceJsonObject



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import Literal, Required, TypedDict

__all__ = ["ToolChoiceString"]


class ToolChoiceString(TypedDict, total=False):
    value: Required[Literal["none", "auto", "required"]]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.


from .._models import BaseModel
from .chat_completion import ChatCompletion

__all__ = ["ChatStreamingData"]


class ChatStreamingData(BaseModel):
    data: ChatCompletion



# This example demonstrates writer.com chat streaming helper
#
# To run it locally you will need:
# - An API key from writer.com
# - Rye installed, see https://rye-up.com
#
# Run the following commands from the root of the repository:
# $ rye sync --all-features
# $ WRITERAI_API_KEY="<your api key>" rye run python examples/chat_streaming_helper.py

from writerai import Client, AsyncClient, BadRequestError


def sync_example() -> None:
    client = Client()

    print("Let's complete a prompt:")
    prompt = "Hi, today I want to write about"
    print(f"> {prompt}")
    try:
        with client.chat.stream(
            model="palmyra-x-004",
            messages=[{"role": "user", "content": prompt}],
        ) as stream:
            for event in stream:
                if event.type == "content.delta":
                    print(event.delta, flush=True, end="")

    except BadRequestError as e:
        print(f"Error: {e.body}")


async def async_example() -> None:
    client = AsyncClient()

    print("Let's complete a prompt:")
    prompt = "Hi, today I want to write about"
    print(f"> {prompt}")
    try:
        async with client.chat.stream(
            model="palmyra-x-004",
            messages=[{"role": "user", "content": prompt}],
        ) as stream:
            async for event in stream:
                if event.type == "content.delta":
                    print(event.delta, flush=True, end="")

    except BadRequestError as e:
        print(f"Error: {e.body}")


if __name__ == "__main__":
    print("=== Synchronous example ===")
    sync_example()
    print()
    print()

    print("=== Asynchronous example ===")
    import asyncio

    asyncio.run(async_example())



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import List
from typing_extensions import Required, TypedDict

__all__ = ["FileRetryParams"]


class FileRetryParams(TypedDict, total=False):
    file_ids: Required[List[str]]
    """The unique identifier of the files to retry."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .._models import BaseModel

__all__ = ["VisionResponse"]


class VisionResponse(BaseModel):
    data: str
    """The result of the image analysis."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing_extensions import Literal

from ..._models import BaseModel

__all__ = ["ToolChoiceString"]


class ToolChoiceString(BaseModel):
    value: Literal["none", "auto", "required"]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Iterable
from typing_extensions import Literal

import httpx

from ..types import vision_analyze_params
from .._types import NOT_GIVEN, Body, Query, Headers, NotGiven
from .._utils import maybe_transform, async_maybe_transform
from .._compat import cached_property
from .._resource import SyncAPIResource, AsyncAPIResource
from .._response import (
    to_raw_response_wrapper,
    to_streamed_response_wrapper,
    async_to_raw_response_wrapper,
    async_to_streamed_response_wrapper,
)
from .._base_client import make_request_options
from ..types.vision_response import VisionResponse

__all__ = ["VisionResource", "AsyncVisionResource"]


class VisionResource(SyncAPIResource):
    @cached_property
    def with_raw_response(self) -> VisionResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/writer/writer-python#accessing-raw-response-data-eg-headers
        """
        return VisionResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> VisionResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/writer/writer-python#with_streaming_response
        """
        return VisionResourceWithStreamingResponse(self)

    def analyze(
        self,
        *,
        model: Literal["palmyra-vision"],
        prompt: str,
        variables: Iterable[vision_analyze_params.Variable],
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> VisionResponse:
        """
        Submit images and a prompt to generate an analysis of the images.

        Args:
          model: The model to use for image analysis.

          prompt: The prompt to use for the image analysis. The prompt must include the name of
              each image variable, surrounded by double curly braces (`{{}}`). For example,
              `Describe the difference between the image {{image_1}} and the image {{image_2}}`.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._post(
            "/v1/vision",
            body=maybe_transform(
                {
                    "model": model,
                    "prompt": prompt,
                    "variables": variables,
                },
                vision_analyze_params.VisionAnalyzeParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=VisionResponse,
        )


class AsyncVisionResource(AsyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AsyncVisionResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/writer/writer-python#accessing-raw-response-data-eg-headers
        """
        return AsyncVisionResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncVisionResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/writer/writer-python#with_streaming_response
        """
        return AsyncVisionResourceWithStreamingResponse(self)

    async def analyze(
        self,
        *,
        model: Literal["palmyra-vision"],
        prompt: str,
        variables: Iterable[vision_analyze_params.Variable],
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> VisionResponse:
        """
        Submit images and a prompt to generate an analysis of the images.

        Args:
          model: The model to use for image analysis.

          prompt: The prompt to use for the image analysis. The prompt must include the name of
              each image variable, surrounded by double curly braces (`{{}}`). For example,
              `Describe the difference between the image {{image_1}} and the image {{image_2}}`.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return await self._post(
            "/v1/vision",
            body=await async_maybe_transform(
                {
                    "model": model,
                    "prompt": prompt,
                    "variables": variables,
                },
                vision_analyze_params.VisionAnalyzeParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=VisionResponse,
        )


class VisionResourceWithRawResponse:
    def __init__(self, vision: VisionResource) -> None:
        self._vision = vision

        self.analyze = to_raw_response_wrapper(
            vision.analyze,
        )


class AsyncVisionResourceWithRawResponse:
    def __init__(self, vision: AsyncVisionResource) -> None:
        self._vision = vision

        self.analyze = async_to_raw_response_wrapper(
            vision.analyze,
        )


class VisionResourceWithStreamingResponse:
    def __init__(self, vision: VisionResource) -> None:
        self._vision = vision

        self.analyze = to_streamed_response_wrapper(
            vision.analyze,
        )


class AsyncVisionResourceWithStreamingResponse:
    def __init__(self, vision: AsyncVisionResource) -> None:
        self._vision = vision

        self.analyze = async_to_streamed_response_wrapper(
            vision.analyze,
        )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import os
from typing import Any, cast

import httpx
import pytest
from respx import MockRouter

from writerai import Writer, AsyncWriter
from tests.utils import assert_matches_type
from writerai.types import (
    File,
    FileRetryResponse,
    FileDeleteResponse,
)
from writerai._response import (
    BinaryAPIResponse,
    AsyncBinaryAPIResponse,
    StreamedBinaryAPIResponse,
    AsyncStreamedBinaryAPIResponse,
)
from writerai.pagination import SyncCursorPage, AsyncCursorPage

base_url = os.environ.get("TEST_API_BASE_URL", "http://127.0.0.1:4010")


class TestFiles:
    parametrize = pytest.mark.parametrize("client", [False, True], indirect=True, ids=["loose", "strict"])

    @parametrize
    def test_method_retrieve(self, client: Writer) -> None:
        file = client.files.retrieve(
            "file_id",
        )
        assert_matches_type(File, file, path=["response"])

    @parametrize
    def test_raw_response_retrieve(self, client: Writer) -> None:
        response = client.files.with_raw_response.retrieve(
            "file_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        file = response.parse()
        assert_matches_type(File, file, path=["response"])

    @parametrize
    def test_streaming_response_retrieve(self, client: Writer) -> None:
        with client.files.with_streaming_response.retrieve(
            "file_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            file = response.parse()
            assert_matches_type(File, file, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_retrieve(self, client: Writer) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `file_id` but received ''"):
            client.files.with_raw_response.retrieve(
                "",
            )

    @parametrize
    def test_method_list(self, client: Writer) -> None:
        file = client.files.list()
        assert_matches_type(SyncCursorPage[File], file, path=["response"])

    @parametrize
    def test_method_list_with_all_params(self, client: Writer) -> None:
        file = client.files.list(
            after="after",
            before="before",
            file_types="file_types",
            graph_id="182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e",
            limit=0,
            order="asc",
            status="in_progress",
        )
        assert_matches_type(SyncCursorPage[File], file, path=["response"])

    @parametrize
    def test_raw_response_list(self, client: Writer) -> None:
        response = client.files.with_raw_response.list()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        file = response.parse()
        assert_matches_type(SyncCursorPage[File], file, path=["response"])

    @parametrize
    def test_streaming_response_list(self, client: Writer) -> None:
        with client.files.with_streaming_response.list() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            file = response.parse()
            assert_matches_type(SyncCursorPage[File], file, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_method_delete(self, client: Writer) -> None:
        file = client.files.delete(
            "file_id",
        )
        assert_matches_type(FileDeleteResponse, file, path=["response"])

    @parametrize
    def test_raw_response_delete(self, client: Writer) -> None:
        response = client.files.with_raw_response.delete(
            "file_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        file = response.parse()
        assert_matches_type(FileDeleteResponse, file, path=["response"])

    @parametrize
    def test_streaming_response_delete(self, client: Writer) -> None:
        with client.files.with_streaming_response.delete(
            "file_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            file = response.parse()
            assert_matches_type(FileDeleteResponse, file, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_delete(self, client: Writer) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `file_id` but received ''"):
            client.files.with_raw_response.delete(
                "",
            )

    @pytest.mark.skip(reason="requests with binary data not yet supported in test environment")
    @parametrize
    @pytest.mark.respx(base_url=base_url)
    def test_method_download(self, client: Writer, respx_mock: MockRouter) -> None:
        respx_mock.get("/v1/files/file_id/download").mock(return_value=httpx.Response(200, json={"foo": "bar"}))
        file = client.files.download(
            "file_id",
        )
        assert file.is_closed
        assert file.json() == {"foo": "bar"}
        assert cast(Any, file.is_closed) is True
        assert isinstance(file, BinaryAPIResponse)

    @pytest.mark.skip(reason="requests with binary data not yet supported in test environment")
    @parametrize
    @pytest.mark.respx(base_url=base_url)
    def test_raw_response_download(self, client: Writer, respx_mock: MockRouter) -> None:
        respx_mock.get("/v1/files/file_id/download").mock(return_value=httpx.Response(200, json={"foo": "bar"}))

        file = client.files.with_raw_response.download(
            "file_id",
        )

        assert file.is_closed is True
        assert file.http_request.headers.get("X-Stainless-Lang") == "python"
        assert file.json() == {"foo": "bar"}
        assert isinstance(file, BinaryAPIResponse)

    @pytest.mark.skip(reason="requests with binary data not yet supported in test environment")
    @parametrize
    @pytest.mark.respx(base_url=base_url)
    def test_streaming_response_download(self, client: Writer, respx_mock: MockRouter) -> None:
        respx_mock.get("/v1/files/file_id/download").mock(return_value=httpx.Response(200, json={"foo": "bar"}))
        with client.files.with_streaming_response.download(
            "file_id",
        ) as file:
            assert not file.is_closed
            assert file.http_request.headers.get("X-Stainless-Lang") == "python"

            assert file.json() == {"foo": "bar"}
            assert cast(Any, file.is_closed) is True
            assert isinstance(file, StreamedBinaryAPIResponse)

        assert cast(Any, file.is_closed) is True

    @pytest.mark.skip(reason="requests with binary data not yet supported in test environment")
    @parametrize
    @pytest.mark.respx(base_url=base_url)
    def test_path_params_download(self, client: Writer) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `file_id` but received ''"):
            client.files.with_raw_response.download(
                "",
            )

    @parametrize
    def test_method_retry(self, client: Writer) -> None:
        file = client.files.retry(
            file_ids=["182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e"],
        )
        assert_matches_type(FileRetryResponse, file, path=["response"])

    @parametrize
    def test_raw_response_retry(self, client: Writer) -> None:
        response = client.files.with_raw_response.retry(
            file_ids=["182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e"],
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        file = response.parse()
        assert_matches_type(FileRetryResponse, file, path=["response"])

    @parametrize
    def test_streaming_response_retry(self, client: Writer) -> None:
        with client.files.with_streaming_response.retry(
            file_ids=["182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e"],
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            file = response.parse()
            assert_matches_type(FileRetryResponse, file, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip(reason="requests with binary data not yet supported in test environment")
    @parametrize
    def test_method_upload(self, client: Writer) -> None:
        file = client.files.upload(
            content=b"raw file contents",
            content_disposition="Content-Disposition",
            content_type="Content-Type",
        )
        assert_matches_type(File, file, path=["response"])

    @pytest.mark.skip(reason="requests with binary data not yet supported in test environment")
    @parametrize
    def test_raw_response_upload(self, client: Writer) -> None:
        response = client.files.with_raw_response.upload(
            content=b"raw file contents",
            content_disposition="Content-Disposition",
            content_type="Content-Type",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        file = response.parse()
        assert_matches_type(File, file, path=["response"])

    @pytest.mark.skip(reason="requests with binary data not yet supported in test environment")
    @parametrize
    def test_streaming_response_upload(self, client: Writer) -> None:
        with client.files.with_streaming_response.upload(
            content=b"raw file contents",
            content_disposition="Content-Disposition",
            content_type="Content-Type",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            file = response.parse()
            assert_matches_type(File, file, path=["response"])

        assert cast(Any, response.is_closed) is True


class TestAsyncFiles:
    parametrize = pytest.mark.parametrize(
        "async_client", [False, True, {"http_client": "aiohttp"}], indirect=True, ids=["loose", "strict", "aiohttp"]
    )

    @parametrize
    async def test_method_retrieve(self, async_client: AsyncWriter) -> None:
        file = await async_client.files.retrieve(
            "file_id",
        )
        assert_matches_type(File, file, path=["response"])

    @parametrize
    async def test_raw_response_retrieve(self, async_client: AsyncWriter) -> None:
        response = await async_client.files.with_raw_response.retrieve(
            "file_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        file = await response.parse()
        assert_matches_type(File, file, path=["response"])

    @parametrize
    async def test_streaming_response_retrieve(self, async_client: AsyncWriter) -> None:
        async with async_client.files.with_streaming_response.retrieve(
            "file_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            file = await response.parse()
            assert_matches_type(File, file, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_retrieve(self, async_client: AsyncWriter) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `file_id` but received ''"):
            await async_client.files.with_raw_response.retrieve(
                "",
            )

    @parametrize
    async def test_method_list(self, async_client: AsyncWriter) -> None:
        file = await async_client.files.list()
        assert_matches_type(AsyncCursorPage[File], file, path=["response"])

    @parametrize
    async def test_method_list_with_all_params(self, async_client: AsyncWriter) -> None:
        file = await async_client.files.list(
            after="after",
            before="before",
            file_types="file_types",
            graph_id="182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e",
            limit=0,
            order="asc",
            status="in_progress",
        )
        assert_matches_type(AsyncCursorPage[File], file, path=["response"])

    @parametrize
    async def test_raw_response_list(self, async_client: AsyncWriter) -> None:
        response = await async_client.files.with_raw_response.list()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        file = await response.parse()
        assert_matches_type(AsyncCursorPage[File], file, path=["response"])

    @parametrize
    async def test_streaming_response_list(self, async_client: AsyncWriter) -> None:
        async with async_client.files.with_streaming_response.list() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            file = await response.parse()
            assert_matches_type(AsyncCursorPage[File], file, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_method_delete(self, async_client: AsyncWriter) -> None:
        file = await async_client.files.delete(
            "file_id",
        )
        assert_matches_type(FileDeleteResponse, file, path=["response"])

    @parametrize
    async def test_raw_response_delete(self, async_client: AsyncWriter) -> None:
        response = await async_client.files.with_raw_response.delete(
            "file_id",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        file = await response.parse()
        assert_matches_type(FileDeleteResponse, file, path=["response"])

    @parametrize
    async def test_streaming_response_delete(self, async_client: AsyncWriter) -> None:
        async with async_client.files.with_streaming_response.delete(
            "file_id",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            file = await response.parse()
            assert_matches_type(FileDeleteResponse, file, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_delete(self, async_client: AsyncWriter) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `file_id` but received ''"):
            await async_client.files.with_raw_response.delete(
                "",
            )

    @pytest.mark.skip(reason="requests with binary data not yet supported in test environment")
    @parametrize
    @pytest.mark.respx(base_url=base_url)
    async def test_method_download(self, async_client: AsyncWriter, respx_mock: MockRouter) -> None:
        respx_mock.get("/v1/files/file_id/download").mock(return_value=httpx.Response(200, json={"foo": "bar"}))
        file = await async_client.files.download(
            "file_id",
        )
        assert file.is_closed
        assert await file.json() == {"foo": "bar"}
        assert cast(Any, file.is_closed) is True
        assert isinstance(file, AsyncBinaryAPIResponse)

    @pytest.mark.skip(reason="requests with binary data not yet supported in test environment")
    @parametrize
    @pytest.mark.respx(base_url=base_url)
    async def test_raw_response_download(self, async_client: AsyncWriter, respx_mock: MockRouter) -> None:
        respx_mock.get("/v1/files/file_id/download").mock(return_value=httpx.Response(200, json={"foo": "bar"}))

        file = await async_client.files.with_raw_response.download(
            "file_id",
        )

        assert file.is_closed is True
        assert file.http_request.headers.get("X-Stainless-Lang") == "python"
        assert await file.json() == {"foo": "bar"}
        assert isinstance(file, AsyncBinaryAPIResponse)

    @pytest.mark.skip(reason="requests with binary data not yet supported in test environment")
    @parametrize
    @pytest.mark.respx(base_url=base_url)
    async def test_streaming_response_download(self, async_client: AsyncWriter, respx_mock: MockRouter) -> None:
        respx_mock.get("/v1/files/file_id/download").mock(return_value=httpx.Response(200, json={"foo": "bar"}))
        async with async_client.files.with_streaming_response.download(
            "file_id",
        ) as file:
            assert not file.is_closed
            assert file.http_request.headers.get("X-Stainless-Lang") == "python"

            assert await file.json() == {"foo": "bar"}
            assert cast(Any, file.is_closed) is True
            assert isinstance(file, AsyncStreamedBinaryAPIResponse)

        assert cast(Any, file.is_closed) is True

    @pytest.mark.skip(reason="requests with binary data not yet supported in test environment")
    @parametrize
    @pytest.mark.respx(base_url=base_url)
    async def test_path_params_download(self, async_client: AsyncWriter) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `file_id` but received ''"):
            await async_client.files.with_raw_response.download(
                "",
            )

    @parametrize
    async def test_method_retry(self, async_client: AsyncWriter) -> None:
        file = await async_client.files.retry(
            file_ids=["182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e"],
        )
        assert_matches_type(FileRetryResponse, file, path=["response"])

    @parametrize
    async def test_raw_response_retry(self, async_client: AsyncWriter) -> None:
        response = await async_client.files.with_raw_response.retry(
            file_ids=["182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e"],
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        file = await response.parse()
        assert_matches_type(FileRetryResponse, file, path=["response"])

    @parametrize
    async def test_streaming_response_retry(self, async_client: AsyncWriter) -> None:
        async with async_client.files.with_streaming_response.retry(
            file_ids=["182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e"],
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            file = await response.parse()
            assert_matches_type(FileRetryResponse, file, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip(reason="requests with binary data not yet supported in test environment")
    @parametrize
    async def test_method_upload(self, async_client: AsyncWriter) -> None:
        file = await async_client.files.upload(
            content=b"raw file contents",
            content_disposition="Content-Disposition",
            content_type="Content-Type",
        )
        assert_matches_type(File, file, path=["response"])

    @pytest.mark.skip(reason="requests with binary data not yet supported in test environment")
    @parametrize
    async def test_raw_response_upload(self, async_client: AsyncWriter) -> None:
        response = await async_client.files.with_raw_response.upload(
            content=b"raw file contents",
            content_disposition="Content-Disposition",
            content_type="Content-Type",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        file = await response.parse()
        assert_matches_type(File, file, path=["response"])

    @pytest.mark.skip(reason="requests with binary data not yet supported in test environment")
    @parametrize
    async def test_streaming_response_upload(self, async_client: AsyncWriter) -> None:
        async with async_client.files.with_streaming_response.upload(
            content=b"raw file contents",
            content_disposition="Content-Disposition",
            content_type="Content-Type",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            file = await response.parse()
            assert_matches_type(File, file, path=["response"])

        assert cast(Any, response.is_closed) is True
